///////////////////////////////////////////////////////////////////////////
// AnsiStr.C - An AVIO ANSI thingy by yours truly.. willit work? I dunno
// Copyright 1991-1993 by John Morris  All rights reserved

#define INCL_SUB
#define INCL_DOSFILEMGR
#include <os2.h>

#include <memory.h>
#include <string.h>

#define ROW        0
#define COLUMN     1

#define TOP        0
#define LEFT       1
#define BOTTOM     2
#define RIGHT      3

#define AND       &&
#define OR        ||
#define NOT        !

typedef USHORT HVPS;
typedef VOID (_fastcall *PVFFN)();

static VOID _fastcall CheckNPositionCursor(SHORT Row, SHORT Column);
static VOID _fastcall ANSIED(VOID);
static VOID _fastcall ANSIEOL(VOID);
static VOID _fastcall ANSICUP(VOID);
static VOID _fastcall ANSICUU(VOID);
static VOID _fastcall ANSICUD(VOID);
static VOID _fastcall ANSICUF(VOID);
static VOID _fastcall ANSICUB(VOID);
static VOID _fastcall ANSISCP(VOID);
static VOID _fastcall ANSIRCP(VOID);
static VOID _fastcall ANSISGR(VOID);
static VOID _fastcall CursorForward(VOID);
static VOID _fastcall CursorDown(VOID);
static VOID _fastcall CursorBackward(VOID);
static VOID _fastcall CursorUp(VOID);
static VOID _fastcall CarriageReturn(VOID);
static VOID _fastcall LineFeed(VOID);
static VOID _fastcall ControlChars(CHAR Character);
static VOID _fastcall DisplayChar(CHAR Character);
static VOID _fastcall StartANSI(VOID);
static VOID _fastcall ProcessANSI(CHAR Character);

typedef struct _ANSITABLE
{
    CHAR    Code;
    PVFFN   Function;

} ANSITABLE;

ANSITABLE ANSITable[] =
{
    'H',    ANSICUP,
    'f',    ANSICUP,
    'A',    ANSICUU,
    'B',    ANSICUD,
    'C',    ANSICUF,
    'D',    ANSICUB,
    's',    ANSISCP,
    'u',    ANSIRCP,
    'J',    ANSIED,
    'K',    ANSIEOL,
    'm',    ANSISGR,
     0,     NULL,
};

static SHORT   WindowSize[4];  // current ANSI window size;
static SHORT   WindowBottomRight[2]; // Bottom and Right side of window
static SHORT   WindowTopLeft[2];     // Top and Left of window

static SHORT   ScreenSize[4];  // Total screen size (may be different than win size)
static SHORT   ScreenBottomRight[2];
static SHORT   ScreenTopLeft[2];

static SHORT   InANSISequence; // whether or not in ANSI sequence
static SHORT   ResetDefault;   // Color reset default holds User.Colors[5]
static BYTE    Color;          // current color attribute

static SHORT   ParmTable[8];   // holds numberic ANSI parms
static SHORT   ParmCount;      // number of parms in 'ParmTable'
static SHORT   ParmTemp;
static SHORT   CursorSave[2];  // used by ANSI save cursor position
static SHORT   CursorAt[2];    // where the cursor is at

static BYTE    ColorTable[8] = { 0, 4, 2, 6, 1, 5, 3, 7 };  // ANSI color table

BOOL    TranslateMode = FALSE;

static HVPS    VideoHvps = 0;  // Video Presentation Space
static VIOMODEINFO VioInfoHold;

// DECLARE SUB InitANSI(BYVAL Rows, BYVAL Columns)

VOID APIENTRY InitANSI(SHORT Rows, SHORT Columns)
{
    VideoHvps = 0;

    VioGetMode(&VioInfoHold, VideoHvps);

    if (Rows > 25 AND Rows < 51)
    {
        VIOMODEINFO VioInfo;

        VioInfo.cb = sizeof(VIOMODEINFO);
        VioGetMode(&VioInfo, VideoHvps);

        VioInfo.row = Rows;

        VioSetMode(&VioInfo, VideoHvps);
    }

    // Init screen & window sizes to full screen

    // Set Top
    ScreenTopLeft[ROW]   = WindowTopLeft[ROW] =
    ScreenSize[TOP]      = WindowSize[TOP] = 0;

    // Set Left
    ScreenTopLeft[COLUMN]= WindowTopLeft[COLUMN] =
    ScreenSize[LEFT]     = WindowSize[LEFT] = 0;

    // Set Bottom
    ScreenBottomRight[ROW]= WindowBottomRight[ROW] =
    ScreenSize[BOTTOM]   = WindowSize[BOTTOM] = Rows - 1;

    // Set Right
    ScreenBottomRight[COLUMN]= WindowBottomRight[COLUMN] =
    ScreenSize[RIGHT]    = WindowSize[RIGHT] = Columns - 1;

    InANSISequence = FALSE;
    ResetDefault = Color = 7;   // white on black..

    memset(ParmTable, 0, 16);

    ParmCount = 0;

    ParmTemp = 0;

    CursorSave[ROW]    = 0;
    CursorSave[COLUMN] = 0;

    return;
}

VOID APIENTRY DeInitANSI(VOID)
{
    VioSetMode(&VioInfoHold, VideoHvps);

    return;
}

// DECLARE SUB SetResetDefault(BYVAL Default)

VOID APIENTRY SetResetDefault(SHORT Default)
{
    ResetDefault = Default;
}

// DECLARE SUB SetANSIWindow(BYVAL Top, BYVAL Left, BYVAL Bottom, BYVAL Right)

VOID APIENTRY SetANSIWindow(SHORT Top, SHORT Left, SHORT Bottom, SHORT Right)
{
    WindowTopLeft[ROW]        = WindowSize[TOP]    = Top - 1;

    WindowTopLeft[COLUMN]     = WindowSize[LEFT]   = Left - 1;

    WindowBottomRight[ROW]    = WindowSize[BOTTOM] = Bottom - 1;

    WindowBottomRight[COLUMN] = WindowSize[RIGHT]  = Right - 1;
}

static VOID _fastcall CheckNPositionCursor(SHORT Row, SHORT Column)
{
    if (Row < WindowTopLeft[ROW])
        Row = WindowTopLeft[ROW];
    else if (Row > WindowBottomRight[ROW])
        Row = WindowBottomRight[ROW];

    if (Column < WindowTopLeft[COLUMN])
        Column = WindowTopLeft[COLUMN];
    else if (Column > WindowBottomRight[COLUMN])
        Column = WindowBottomRight[COLUMN];

    CursorAt[ROW]    = Row;
    CursorAt[COLUMN] = Column;

    VioSetCurPos(Row, Column, VideoHvps);
}

static VOID _fastcall ANSIED(VOID)
{
    BYTE CLSCell[2];

    CLSCell[0] = ' ';
    CLSCell[1] = Color;

    if (ParmTable[0] == 2)
    {
        VioScrollUp(WindowSize[TOP],    WindowSize[LEFT],
                    WindowSize[BOTTOM], WindowSize[RIGHT],
                    0xFFFF, CLSCell, VideoHvps);

        CursorAt[ROW] = 0;
        CursorAt[COLUMN] = 0;

        VioSetCurPos(CursorAt[ROW], CursorAt[COLUMN], VideoHvps);
    }

    // This area here would be for erasing to the end of the line, and to
    // the bottom of the screen.. for later usage..
}

static VOID _fastcall ANSIEOL(VOID)
{
    BYTE EOLCell[2];

    if (ParmCount)
        return;

    EOLCell[0] = ' ';
    EOLCell[1] = Color;

    VioWrtNCell(EOLCell, (WindowSize[RIGHT] - CursorAt[COLUMN]) + 1,
                CursorAt[ROW], CursorAt[COLUMN], VideoHvps);
}

static VOID _fastcall ANSICUP(VOID)
{
    if (ParmCount < 1)
        ParmTable[0] = 1;
    if (ParmCount < 2)
        ParmTable[1] = 1;

    CheckNPositionCursor(ParmTable[0] - 1, ParmTable[1] - 1);
}

static VOID _fastcall ANSICUU(VOID)
{
    SHORT Row = CursorAt[ROW];

    if (ParmCount)
        Row -= ParmTable[0];
    else
        Row--;

    CheckNPositionCursor(Row, CursorAt[COLUMN]);
}

static VOID _fastcall ANSICUD(VOID)
{
    SHORT Row = CursorAt[ROW];

    if (ParmCount)
        Row += ParmTable[0];
    else
        Row++;

    CheckNPositionCursor(Row, CursorAt[COLUMN]);
}

static VOID _fastcall ANSICUF(VOID)
{
    SHORT Column = CursorAt[COLUMN];

    if (ParmCount)
        Column += ParmTable[0];
    else
        Column++;

    CheckNPositionCursor(CursorAt[ROW], Column);
}

static VOID _fastcall ANSICUB(VOID)
{
    SHORT Column = CursorAt[COLUMN];

    if (ParmCount)
        Column -= ParmTable[0];
    else
        Column--;

    CheckNPositionCursor(CursorAt[ROW], Column);
}

static VOID _fastcall ANSISCP(VOID)
{
    CursorSave[ROW]    = CursorAt[ROW];
    CursorSave[COLUMN] = CursorAt[COLUMN];
}

static VOID _fastcall ANSIRCP(VOID)
{
    CursorAt[ROW]    = CursorSave[ROW];
    CursorAt[COLUMN] = CursorSave[COLUMN];

    CheckNPositionCursor(CursorAt[ROW], CursorAt[COLUMN]);
}

static VOID _fastcall ANSISGR(VOID)
{
    SHORT   Index;
    BYTE    TempFG, TempBG;

    if (NOT ParmCount)
    {
        ParmCount = 1;
        ParmTable[0] = 0;
        ParmTable[1] = ResetDefault;
    }

    for (Index = 0; Index < ParmCount; Index++)
    {
        if (ParmTable[Index] < 9)   // Attribute thing
        {
            switch (ParmTable[Index])
            {
                case 0:
                    Color = 7;
                    break;
                case 1:
                    Color |= 0x08;   // intensity bit
                    break;
                case 5:
                    Color |= 0x80;   // blink bit
                    break;
                case 7:
                    TempFG = (BYTE) (Color & 0x0F);
                    TempBG = (BYTE) (Color >> 4);
                    Color  = (BYTE) (TempFG << 4) + TempBG;
                    break;
                case 8:
                    TempFG = (BYTE) (Color >> 4);
                    Color  = (BYTE) (Color & 0xF0) + TempFG;
                    break;
            }
        }
        else if (ParmTable[Index] > 29 AND ParmTable[Index] < 38)
        {
            Color = (BYTE) (Color & 0xF8) + ColorTable[ParmTable[Index] - 30];
        }
        else if (ParmTable[Index] > 39 AND ParmTable[Index] < 48)
        {
            Color = (BYTE) (Color & 0x8F) + (ColorTable[ParmTable[Index] - 40] << 4);
        }
    }
}

BOOL   ForcedLF = FALSE;
USHORT LAdded;

static VOID _fastcall CursorForward(VOID)
{
    if (CursorAt[COLUMN] == WindowBottomRight[COLUMN])
    {
        ForcedLF = TRUE;

        LAdded++;

        CursorAt[COLUMN] = WindowTopLeft[COLUMN];

        CursorDown();
    }
    else
        CursorAt[COLUMN]++;
}

static VOID _fastcall CursorDown(VOID)
{
    BYTE ScrollCell[2];

    ScrollCell[0] = ' ';
    ScrollCell[1] = Color;

    if (CursorAt[ROW] == WindowBottomRight[ROW])
    {
        VioScrollUp(WindowSize[TOP], WindowSize[LEFT],
                    WindowSize[BOTTOM], WindowSize[RIGHT],
                    1, ScrollCell, VideoHvps);
    }
    else
        CursorAt[ROW]++;
}

static VOID _fastcall CursorBackward(VOID)
{
    if (CursorAt[COLUMN] == WindowTopLeft[COLUMN])
    {
        if (CursorAt[ROW] > WindowTopLeft[ROW])
        {
            CursorAt[COLUMN] = WindowBottomRight[COLUMN];

            CursorUp();
        }
    }
    else
        CursorAt[COLUMN]--;
}

static VOID _fastcall CursorUp(VOID)
{
    if (CursorAt[ROW] > WindowTopLeft[ROW])
        CursorAt[ROW]--;
}

static VOID _fastcall CarriageReturn(VOID)
{
    CursorAt[COLUMN] = WindowTopLeft[COLUMN];
}

static VOID _fastcall LineFeed(VOID)
{
    if (TranslateMode)
        CarriageReturn();

    if (NOT ForcedLF)
        CursorDown();
}

static VOID _fastcall ControlChars(CHAR Character)
{
    SHORT Index;
    BYTE  CharCell[2];

    switch (Character)
    {
        case 7:
            DosBeep(500, 100);
            break;
        case 8:
            CursorBackward();
            break;
        case 9:
            Character = ' ';

            for (Index = 0; Index < 4; Index++)
                DisplayChar(Character);

            break;
        case 10:
            LineFeed();
            break;
        case 12:
            CharCell[0] = ' ';
            CharCell[1] = Color;

            VioScrollUp(WindowSize[TOP], WindowSize[LEFT],
                        WindowSize[BOTTOM], WindowSize[RIGHT],
                        0xFFFF, CharCell, VideoHvps);
            break;
        case 13:
            CarriageReturn();
            break;
        default:
            DisplayChar(Character);
    }
}

static VOID _fastcall DisplayChar(CHAR Character)
{
    BYTE CharCell[2];

    CharCell[0] = Character;
    CharCell[1] = Color;

    VioWrtNCell(CharCell, 1, CursorAt[ROW], CursorAt[COLUMN], VideoHvps);

    CursorForward();
}

static VOID _fastcall StartANSI(VOID)
{
    memset(ParmTable, 0, 16);

    ParmCount = 0;

    ParmTemp = 0;

    InANSISequence = 1;
}

static VOID _fastcall ProcessANSI(CHAR Character)
{
    if (InANSISequence == 1 AND Character == '[')
    {
        InANSISequence = 2;

        return;
    }
    else
    {
        if (Character == ';')
        {
            if (ParmCount < 8)
            {
                ParmTable[ParmCount++] = ParmTemp;

                ParmTemp = 0;
            }

            return;
        }

        if (Character < '0')
        {
CancelANSI:
            InANSISequence = FALSE;

            DisplayChar(27);

            DisplayChar(Character);

            return;
        }

        if (Character >  '9')
        {
            SHORT Index = 0;

            InANSISequence = -1;

            while (ANSITable[Index].Code)
            {
                if (ANSITable[Index].Code == Character)
                    break;

                Index++;
            }

            if (ANSITable[Index].Code)
            {
                if (ParmTemp)
                    ParmTable[ParmCount++] = ParmTemp;

                (ANSITable[Index].Function)();
            }
            else
                goto CancelANSI;

            return;
        }

        Character -= '0';

        ParmTemp = (ParmTemp * 10) + Character;
    }
}

// Basic's DECLARE SUB ANSIStr(SEG String$, BYVAL Length, SEG Row, SEG Column, SEG LinesAdded)

extern void far pascal StringAssign(PSZ Source, SHORT SLen, PSZ Dest, SHORT DLen );

UCHAR String[513];

VOID APIENTRY ANSIStr(PSZ String16, USHORT Length, PUSHORT Row, PUSHORT Col, PUSHORT LinesAdded)
{
    USHORT  Index;
    VIOCURSORINFO Cursor;

    StringAssign(String16, 0, String, 512);

    LAdded = 0;

    VioGetCurType(&Cursor, VideoHvps);
    Cursor.attr = -1;
    VioSetCurType(&Cursor, VideoHvps);

    for (Index = 0; Index < Length; Index++)
    {
        if (String[Index] == 10 OR String[Index] == 13)
        {
            ControlChars(String[Index]);

            continue;
        }

        ForcedLF = FALSE;

        if (InANSISequence == -1)
            InANSISequence = FALSE;

        if (NOT InANSISequence)
        {
            if (String[Index] < ' ')
            {
                if (String[Index] == 27)
                    StartANSI();
                else
                    ControlChars(String[Index]);
            }
            else
                DisplayChar(String[Index]);
        }
        else
            ProcessANSI(String[Index]);
    }

    VioSetCurPos(CursorAt[ROW], CursorAt[COLUMN], VideoHvps);

    Cursor.attr = 0;
    VioSetCurType(&Cursor, VideoHvps);

    VioGetCurPos(Row, Col, 0);

    (*Row)++;
    (*Col)++;
    *LinesAdded = LAdded;
    return;
}

// DECLARE SUB ANSIStrAt(SEG String16, BYVAL Length AS INTEGER, BYVAL Row AS INTEGER, BYVAL Col AS INTEGER, DIM Attr AS STRING * 1)

VOID APIENTRY ANSIStrAt(PSZ String16, USHORT Length, USHORT Row, USHORT Col, USHORT Attr)
{
    BYTE Attribute = (BYTE) Attr;

    StringAssign(String16, 0, String, 512);

    VioWrtCharStrAtt(String, Length, Row - 1, Col - 1, &Attribute, VideoHvps);

    return;
}

//  DECLARE SUB ANSIStrTTY(SEG String16$, BYVAL Length AS INTEGER)

VOID APIENTRY ANSIStrTTY(PSZ String16, USHORT Length)
{
    StringAssign(String16, 0, String, 512);
                   
    VioWrtTTY(String, strlen(String), VideoHvps);

    return;
}

// DECLARE SUB ANSICurLoc(BYVAL Row AS INTEGER, BYVAL Col AS INTEGER)

VOID APIENTRY ANSICurLoc(USHORT Row, USHORT Col)
{
    CheckNPositionCursor(Row - 1, Col - 1);

    return;
}

CHAR TempReadBuf[129];

// DECLARE FUNCTION ReadCom%(BYVAL Handle, SEG String16$, BYVAL Bytes)

USHORT APIENTRY ReadCom(HFILE Handle, PSZ String16, USHORT Bytes)
{
    USHORT ReturnCode;
    USHORT BytesRead;

    ReturnCode = DosRead(Handle, &TempReadBuf, Bytes, &BytesRead);

    StringAssign(TempReadBuf, BytesRead, String16, 0);

    return (ReturnCode);
}

